@@ -8,8 +8,8 @@ function (formula, data = NULL, ..., subset, na.action = na.omit, scale = TRUE)
   if (!inherits(formula, "formula")) 
     stop("method is only for formula objects")
   m <- match.call(expand.dots = FALSE)
-  if (is.matrix(eval(m$data, parent.frame())))
-    m$data <- as.data.frame(data)
+  if (identical(class(eval.parent(m$data)), "matrix"))
+    m$data <- as.data.frame(eval.parent(m$data))
   m$... <- NULL
   m$scale <- NULL
   m[[1]] <- as.name("model.frame")
@@ -19,6 +19,7 @@ function (formula, data = NULL, ..., subset, na.action = na.omit, scale = TRUE)
   attr(Terms, "intercept") <- 0
   x <- model.matrix(Terms, m)
   y <- model.extract(m, response)
+  attr(x, "na.action") <- attr(y, "na.action") <- attr(m, "na.action")
   if (length(scale) == 1)
     scale <- rep(scale, ncol(x))
   if (any(scale)) {
@@ -28,7 +29,7 @@ function (formula, data = NULL, ..., subset, na.action = na.omit, scale = TRUE)
                      )
     scale <- !attr(x, "assign") %in% remove
   }
-  ret <- svm.default (x, y, scale = scale, ...)
+  ret <- svm.default (x, y, scale = scale, ..., na.action = na.action)
   ret$call <- call
   ret$call[[1]] <- as.name("svm")
   ret$terms <- Terms
@@ -118,21 +119,24 @@ function (x,
         df <- na.action(data.frame(y, x))
         y <- df[,1]
         x <- as.matrix(df[,-1])
+        attr(x, "na.action") <- attr(y, "na.action") <- attr(df, "na.action")
       }
     }
 
+    nac <- attr(x, "na.action")
+    
     ## scaling
     if (length(scale) == 1)
       scale <- rep(scale, ncol(x))
     if (any(scale)) {
       co <- !apply(x[,scale, drop = FALSE], 2, var)
       if (any(co)) {
-        scale <- rep(FALSE, ncol(x))
         warning(paste("Variable(s)",
-                      paste("`",colnames(x[,scale, drop = FALSE])[co],
-                            "'", sep="", collapse=" and "),
+                      paste(sQuote(colnames(x[,scale, drop = FALSE])[co]),
+                            sep="", collapse=" and "),
                       "constant. Cannot scale data.")
                 )
+        scale <- rep(FALSE, ncol(x))
       } else {
         xtmp <- scale(x[,scale])
         x[,scale] <- xtmp
@@ -275,7 +279,8 @@ function (x,
                coefs    = if (cret$nr == 0) NULL else
                               t(matrix(cret$coefs[1:((cret$nclasses - 1) * cret$nr)],
                                        nrow = cret$nclasses - 1,
-                                       byrow = TRUE))
+                                       byrow = TRUE)),
+               na.action = nac
               )
 
   # cross-validation-results
@@ -291,7 +296,12 @@ function (x,
     }
 
   class (ret) <- "svm"
-  ret$fitted  <- if (fitted) predict(ret, xhold) else NA
+
+  if (fitted) {
+    ret$fitted <- na.action(predict(ret, xhold))
+    if (type > 1) ret$residuals <- y - ret$fitted
+  }
+  
   ret
 } 
 
@@ -303,22 +313,39 @@ predict.svm <- function (object, newdata,
   if (missing(newdata))
     return(fitted(object))
 
+  if (object$tot.nSV < 1)
+    stop("Model is empty!")
+  
   sparse <- inherits(newdata, "matrix.csr")
   if (object$sparse || sparse) {
     if (!require(SparseM))
       stop("Need SparseM package for handling of sparse structures!")
   }
 
+  act <- NULL
   if (is.vector(newdata) || sparse) newdata <- t(t(newdata))
+  preprocessed <- !is.null(attr(newdata, "na.action"))
+  rowns <- if (!is.null(rownames(newdata)))
+    rownames(newdata)
+  else
+    1:nrow(newdata)
   if (!object$sparse) {
     if (inherits(object, "svm.formula")) {
       if(is.null(colnames(newdata)))
         colnames(newdata) <- colnames(object$SV)
+      newdata <- na.action(newdata)
+      act <- attr(newdata, "na.action")
       newdata <- model.matrix(delete.response(terms(object)),
                               as.data.frame(newdata), na.action = na.action)
-    } else if (!sparse) newdata <- na.action(as.matrix(newdata))
+    } else if (!sparse) {
+      newdata <- na.action(as.matrix(newdata))
+      act <- attr(newdata, "na.action")
+    }
   }
 
+  if (!is.null(act) && !preprocessed)
+    rowns <- rowns[-act]
+  
   if (any(object$scaled))
     newdata[,object$scaled] <-
       scale(newdata[,object$scaled, drop = FALSE],
@@ -369,33 +396,38 @@ predict.svm <- function (object, newdata,
 
              PACKAGE = "e1071"
             )
-  
-  ret2 <- if (is.character(object$levels))
-    # classification: return factors
+
+  ret2 <- if (is.character(object$levels)) # classification: return factors
     factor (object$levels[ret$ret], levels = object$levels)
-  else if (object$type == 2)
-    # one-class-classification: return TRUE/FALSE
+  else if (object$type == 2) # one-class-classification: return TRUE/FALSE
     ret$ret == 1 
-  else if (any(object$scaled))
-    # return raw values, possibly scaled back
+  else if (any(object$scaled)) # return raw values, possibly scaled back
     ret$ret * object$y.scale$"scaled:scale" + object$y.scale$"scaled:center"
   else
     ret$ret
 
+  names(ret2) <- rowns
+  ret2 <- napredict(act, ret2)
+  
   if (decision.values) {
     colns = c()
     for (i in 1:(object$nclasses - 1))
       for (j in (i + 1):object$nclasses)
         colns <- c(colns, paste(object$levels[object$labels[i]],
                                 "/", object$levels[object$labels[j]], sep = ""))
-    attr(ret2, "decision.values") <- matrix(ret$dec, nrow = nrow(newdata), byrow = TRUE)
-    colnames(attr(ret2, "decision.values")) <- colns
+    attr(ret2, "decision.values") <- napredict(act,
+                                               matrix(ret$dec, nrow = nrow(newdata), byrow = TRUE,
+                                                      dimnames = list(rowns, colns)
+                                                      )
+                                               )
   }
 
-  if (probability && object$type < 2) {
-    attr(ret2, "probabilities") <- matrix(ret$prob, nrow = nrow(newdata), byrow = TRUE)
-    colnames(attr(ret2, "probabilities")) <- object$levels[object$labels]
-  }
+  if (probability && object$type < 2)
+    attr(ret2, "probabilities") <- napredict(act,
+                                             matrix(ret$prob, nrow = nrow(newdata), byrow = TRUE,
+                                                    dimnames = list(rowns, colns)
+                                                    )
+                                             )
 
   ret2
 }
@@ -508,7 +540,7 @@ plot.svm <- function(x, data, formula = NULL, fill = TRUE,
                      levels = 1:(length(unique(as.numeric(preds)))+1),
                      key.axes = axis(4,
                        1:length(unique(as.numeric(preds)))+0.5,
-                       labels = levels(preds)[unique(preds)], las = 3
+                       labels = levels(preds), las = 3
                        ),
                      plot.title = title(main = "SVM classification plot",
                        xlab = names(lis)[2], ylab = names(lis)[1]),
